



if not exist %LIBRARY_INC%\gudhi\Simplex_tree.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_alpha_complex.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_bottleneck_distance.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_cubical_complex.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_euclidean_witness_complex.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_reader_utils.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_rips_complex.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_simplex_generators.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_simplex_tree.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_subsampling.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_tangential_complex.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_time_delay.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest src/python/test/test_witness_complex.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
